# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("game", "0064_character_name_data")]

    operations = [
        migrations.RenameField(
            model_name="level", old_name="character", new_name="character_old"
        )
    ]



from django.db import migrations


def update_level(apps, schema_editor):

    Level = apps.get_model("game", "Level")
    level74 = Level.objects.get(name="74", default=1)
    level74.model_solution = "[14]"
    level74.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0073_level_75_solution")]
    operations = [
        migrations.RunPython(update_level, reverse_code=migrations.RunPython.noop)
    ]



import json

from common.tests.utils.classes import create_class_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from deploy import captcha
from django.test.client import Client
from django.test.testcases import TestCase
from django.urls import reverse
from hamcrest import *

from .test_level_editor import LevelEditorTestCase
from .utils.level import create_save_level


class LevelModerationTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.orig_captcha_enabled = captcha.CAPTCHA_ENABLED
        captcha.CAPTCHA_ENABLED = False
        super(LevelModerationTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        captcha.CAPTCHA_ENABLED = cls.orig_captcha_enabled
        super(LevelModerationTestCase, cls).tearDownClass()

    def setUp(self):
        self.client = Client()

    def test_moderation_teachers_class(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)

        level_id = create_save_level(student, level_name)

        self.teacher_login(email, password)

        level_moderation_url = reverse("level_moderation")
        response = self.client.get(level_moderation_url)
        assert class_name in response.content.decode()
        assert level_name in response.content.decode()

        # Test delete level
        delete_level_url = reverse("delete_level", kwargs={"levelID": level_id})
        self.client.post(delete_level_url)
        response = self.client.get(level_moderation_url)
        assert level_name not in response.content.decode()

    def test_moderation_another_class(self):
        level_name = "test_level2"
        email, password = signup_teacher_directly()

        email2, _ = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email2)

        _, _, student = create_school_student_directly(access_code)

        create_save_level(student, level_name)

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.get(url)
        assert class_name not in response.content.decode()
        assert level_name not in response.content.decode()

    def test_moderation_empty_class_filter(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)

        create_save_level(student, level_name)

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.post(url, {"classes": []})
        assert class_name in response.content.decode()
        assert level_name not in response.content.decode()
        assert "No levels found." in response.content.decode()

    def test_moderation_shared_with(self):
        level_name = "test_level1"
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email)

        _, _, student = create_school_student_directly(access_code)
        _, _, student2 = create_school_student_directly(access_code)

        create_save_level(student, level_name, shared_with=[student2.new_user])

        self.teacher_login(email, password)

        url = reverse("level_moderation")
        response = self.client.get(url)
        assert class_name in response.content.decode()
        assert level_name in response.content.decode()
        assert student2.new_user.first_name in response.content.decode()

    def teacher_login(self, email, password):
        self.client.post(
            reverse("teacher_login"),
            {
                "auth-username": email,
                "auth-password": password,
                "teacher_login_view-current_step": "auth",
            },
            follow=True,
        )



from game.end_to_end_tests.base_game_test import BaseGameTest
from game.views.level_editor import available_blocks


class TestLevelEditor(BaseGameTest):
    def test_level_editor_displays(self):
        page = self.go_to_level_editor()

        assert page.element_exists_by_id("paper")

    def test_code_tab_blocks_load(self):
        page = self.go_to_level_editor()
        page.go_to_code_tab()

        block_types = [block.type for block in available_blocks()]

        for block_type in block_types:
            assert page.element_exists_by_id(f"{block_type}_checkbox")
            assert page.element_exists_by_id(f"{block_type}_image")



from __future__ import absolute_import
from __future__ import division

from common.models import Student, Class
from django.shortcuts import render
from portal.templatetags import app_tags

import game.messages as messages
import game.permissions as permissions
from game.forms import LevelModerationForm
from game.models import Level
from .helper import renderError


def level_moderation(request):
    """Renders a page with students' scores.

    **Context**

    ``RequestContext``
    ``form``
        Form used to choose a class and level to show. Instance of `forms.ScoreboardForm.`
    ``studentData``
        List of lists containing all the data to be stored in the scoreboard table.
    ``thead``
        List of Strings representing the headers of the scoreboard table.

    **Template:**

    :template:`game/level_moderation.html`
    """

    # Not showing this part to outsiders.
    if not permissions.can_see_level_moderation(request.user):
        return renderError(
            request,
            messages.noPermissionLevelModerationTitle(),
            messages.noPermissionLevelModerationPage(),
        )

    teacher = request.user.userprofile.teacher
    classes_taught = Class.objects.filter(teacher=teacher)

    if len(classes_taught) <= 0:
        return renderError(
            request,
            messages.noPermissionLevelModerationTitle(),
            messages.noDataToShowLevelModeration(),
        )

    classes_taught_ids = [class_.id for class_ in classes_taught]
    form = LevelModerationForm(
        request.POST or None,
        classes=classes_taught,
        initial={"classes": classes_taught_ids},
    )

    if request.method == "POST":
        if form.is_valid():
            class_ids = set(map(int, request.POST.getlist("classes")))

            # check user has permission to look at the classes
            if not all(class_id in classes_taught_ids for class_id in class_ids):
                return renderError(
                    request,
                    messages.noPermissionLevelModerationTitle(),
                    messages.noPermissionLevelModerationClass(),
                )
        else:
            class_ids = []
    else:
        class_ids = [class_id for class_id in classes_taught_ids]

    students = Student.objects.filter(
        class_field_id__in=class_ids, new_user__is_active=True
    )
    owners = [student.user for student in students]

    table_headers = [
        "Student",
        "Level name",
        "Shared with",
        "Actions",
    ]
    level_data = []

    for owner in owners:
        for level in Level.objects.filter(owner=owner):
            users_shared_with = [
                user
                for user in level.shared_with.all()
                if permissions.CanShareLevelWith().can_share_level_with(
                    user, owner.user
                )
                and user != owner.user
            ]

            if not users_shared_with:
                shared_str = "-"
            else:
                shared_str = ", ".join(
                    app_tags.make_into_username(user) for user in users_shared_with
                )

            level_data.append(
                {
                    "student": app_tags.make_into_username(owner.user),
                    "id": level.id,
                    "name": level.name,
                    "shared_with": shared_str,
                }
            )

    return render(
        request,
        "game/level_moderation.html",
        context={
            "form": form,
            "levelData": level_data,
            "thead": table_headers,
        },
    )



"""CSP config"""

from common.app_settings import domain

CSP_CONFIG = {
    "CSP_DEFAULT_SRC": ("'self'",),
    "CSP_IMG_SRC": (
        f"{domain()}/static/game/image/",
        f"{domain()}/static/game/raphael_image/",
        f"{domain()}/static/game/js/blockly/media/",
        f"{domain()}/static/icons/",
        "https://cdn.crowdin.com/",
        "https://crowdin-static.downloads.crowdin.com/"
    ),
    "CSP_STYLE_SRC": ("https://code.jquery.com/ui/1.13.1/themes/base/jquery-ui.css", "https://cdn.crowdin.com/"),
    "CSP_SCRIPT_SRC": ("'unsafe-eval'", "https://cdn.crowdin.com/"),
    "CSP_FRAME_SRC": (
        f"{domain()}/static/game/image/",
        "https://crowdin.com/"
    ),
    "CSP_OBJECT_SRC": (f"{domain()}/static/game/image/",),
    "CSP_CONNECT_SRC": ("https://crowdin.com/",)
}



import pytest


@pytest.mark.django_db
def test_episodes_renamed_properly(migrator):
    old_state = migrator.apply_initial_migration(("game", "0067_level_score_27"))
    new_state = migrator.apply_tested_migration(
        ("game", "0069_remove_user_levels_from_episodes")
    )

    Episode = new_state.apps.get_model("game", "Episode")
    episode7 = Episode.objects.get(id=7)
    episode8 = Episode.objects.get(id=8)

    assert episode7.name == "Limited Blocks"
    assert episode8.name == "Procedures"


@pytest.mark.django_db
def test_episodes_reordered_properly(migrator):
    old_state = migrator.apply_initial_migration(("game", "0067_level_score_27"))
    new_state = migrator.apply_tested_migration(
        ("game", "0069_remove_user_levels_from_episodes")
    )

    Episode = new_state.apps.get_model("game", "Episode")
    episode6 = Episode.objects.get(id=6)
    episode7 = Episode.objects.get(id=7)
    episode8 = Episode.objects.get(id=8)
    episode9 = Episode.objects.get(id=9)

    assert episode7.next_episode == episode8
    assert episode8.next_episode == episode9
    assert episode6.next_episode == episode7



from builtins import str


def level_creation_email_subject():
    return "Custom level to moderate"


def level_creation_email_text_content():
    return (
        "Your student {student_name} from your class {class_name} has created this level: {level_url}. "
        "If you want to moderate it, you can do it on the moderation board: {moderate_url}."
    )


def youtubeLink(width, height, url, border):
    return (
        "<iframe width='"
        + str(width)
        + "' height='"
        + str(height)
        + "' src='"
        + str(url)
        + "?rel=0"
        + "' frameborder='"
        + str(border)
        + "' allowfullscreen class='video'></iframe><br>"
    )


def play_button_icon_url():
    return (
        '<img src="/static/game/image/actions/go.svg" alt="Play button" '
        'style="width: 4%;">'
    )


def step_button_icon_url():
    return (
        '<img src="/static/game/image/icons/step.svg" alt="Step button" '
        '" style="width: 4%;">'
    )


def noPermissionMessage():
    return "You have no permission to see this."


def notSharedLevel():
    return (
        "This level is private. You can only see the public levels and the ones "
        "created by other users only if they share them with you."
    )


""" Strings used in the scoreboard. """


def noPermissionTitle():
    return "No permission"


def noPermissionScoreboard():
    return (
        "Scoreboard is only visible to school students and teachers. Log in if you "
        "think you should be able to see it."
    )


def noDataToShow():
    return (
        "There is no data to show. Please contact your administrator if this is "
        "unexpected."
    )


""" Strings used on the level moderation page. """


def noPermissionLevelModerationTitle():
    return "No permission"


def noPermissionLevelModerationPage():
    return (
        "Level moderation is only visible to teachers. Log in if you "
        "think you should be able to see it."
    )


def noPermissionLevelModerationClass():
    return (
        "You do not teach this class. Please contact your administrator if this "
        "is unexpected."
    )


def noPermissionLevelModerationStudent():
    return (
        "You do not teach this student. Please contact your administrator if this "
        "is unexpected."
    )


def noDataToShowLevelModeration():
    return (
        "You have not created any classes and therefore is no data to show. "
        "Please contact your administrator if this is unexpected."
    )


""" String messages used on the settings page. """


def shareTitle():
    return "Level Share"


def shareSuccessfulPerson(name, surname):
    return f"You shared your level with {name} {surname} successfully!"


def shareSuccessfulClass(className):
    return f"You shared your level with class {className} successfully!"


def shareUnsuccessfulPerson(first_name, last_name):
    return (
        f"We were unable to find {first_name} {last_name}. Are you sure you got their "
        f"name right?"
    )


def shareUnsuccessfulClass(className):
    return f"We were unable to find class {className}. Are you sure you got it right?"


def noLevelsToShow():
    return "It seems that you have not created any levels. How about creating one now?"


def levelsMessage():
    return (
        "All the levels you have created so far. Click on them to play them or share "
        "them with your friends."
    )


def sharedLevelsMessage():
    return (
        "All the levels created by others that were shared with you. Click on them to "
        "play them."
    )


def noSharedLevels():
    return "No one shared a level with you yet."


""" Strings used in the class view. """


def chooseClass():
    return "Choose a class you want to see."


def noPermission():
    return "You don't have permissions to see this."


""" String messages used as level tips in the game view. """


def title_night_mode():
    return "Can you find your way in the dark?"


def build_description(title, message):
    return f"<b>{title}</b><br><br>{message}"


def title_level_default():
    return " "


def description_level_default():
    return "Can you find the shortest route?"


def hint_level_default():
    return "Think back to earlier levels. What did you learn?"


def title_level1():
    return "Can you help the van get to the house?"


def description_level1():
    message = (
        f"Choose the right blocks to tell the van where to go. <br> Drag the "
        f"blocks under the <b>Start</b> block to attach them. <br> To remove a "
        f"block, drag it into the bin in the bottom right of the screen. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level1(), message)


def hint_level1():
    return (
        "Try dragging a move forwards block onto the <b>Start</b> block. Then click "
        "the <b>Start</b> block."
    )


def title_level2():
    return "This time the house is further away"


def description_level2():
    message = (
        f"A block can be placed next to or under another, like a jigsaw. A second "
        f"<b>Move forwards</b> block can be placed under the first <b>Move "
        f"forwards</b> block. <br> To remove a block, drag it back to the "
        f"left of the screen or drop it in the bin. <br> When you are happy with "
        f"your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level2(), message)


def hint_level2():
    return (
        "Now there is further to go. How many times do you need to move forwards this "
        "time?"
    )


def title_level3():
    return "Can you make the van turn right?"


def description_level3():
    message = (
        f"This time, the van has to turn right to reach the house. Make sure you use "
        f"the <b>Turn right</b> block in your sequence. <br> Drag the blocks "
        f"and attach them under the <b>Start</b> block like before. To remove a "
        f"block, drag it back to the left of the screen or drop it in the bin. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level3(), message)


def hint_level3():
    return "Now you need to make a turn. Which way do you need to go?"


def title_level4():
    return "You are getting good at this! Let's try turning left"


def description_level4():
    message = (
        f"This time the van has to go left. Make sure you use the <b>Turn left</b> "
        f"block in your sequence. <br> Drag and attach the blocks like before. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level4(), message)


def hint_level4():
    return (
        "You’re doing really well! This time you need to go a bit further and turn the "
        "other way. Count the number of blocks you need to move."
    )


def title_level5():
    return "Good work! You are ready for something harder"


def description_level5():
    message = (
        f"You already know how to make the van turn left or right. This time "
        f"the van has to make lots of turns to reach the house. <br> Drag and "
        f"attach the blocks to make your sequence."
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level5(), message)


def hint_level5():
    return (
        "There are more turns to make this time. Try building your solution up block "
        "by block. Which way do you need to turn first?"
    )


def title_level6():
    return "Well done! Let's use all three blocks"


def description_level6():
    message = (
        f"This time the van has to <b>Move forwards</b>, <b>Turn left</b> and "
        f"<b>Turn right</b>. <br><br> Drag and attach the blocks like before. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level6(), message)


def hint_level6():
    return (
        "You’ve done really well to get this far! Now you need to travel further but "
        "include some turns. Check which direction you need to turn and count the "
        "number of blocks you need to travel."
    )


def title_level7():
    return "This road is more complicated"


def description_level7():
    message = (
        f"Practise your new skills on this road by helping the driver to arrive at "
        f"the house. <br> When you are happy with your sequence, press "
        f"{play_button_icon_url()}!"
    )
    return build_description(title_level7(), message)


def hint_level7():
    return (
        "This road is more complicated but you can do this! Start with just a few "
        "blocks and try that, then add more blocks and check it again."
    )


def title_level8():
    return "The warehouse is not always in the same place"


def description_level8():
    message = (
        f"This time the warehouse is somewhere else on the screen but you still need "
        f"to use the <b>Move forwards</b> block. <br> Can you use the <b>Move "
        f"forwards</b> block correctly even when it looks like the van goes in a "
        f"different direction? "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level8(), message)


def hint_level8():
    return (
        "The warehouse is in a different place now but don’t let that confuse you! "
        "Which way do you need to go first? Imagine you are driving the van."
    )


def title_level9():
    return "Can you go from right to left?"


def description_level9():
    message = (
        f"Practise your new skills on this road by helping the driver to arrive "
        f"at the house. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level9(), message)


def hint_level9():
    return "Try counting the blocks before the turn. Which way do you need to go?"


def title_level10():
    return "Well done! How about another go?"


def description_level10():
    message = (
        f"You've done really well so far. Try to get the van to the house. "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level10(), message)


def hint_level10():
    return (
        "You need to make a turn as soon as you leave the warehouse. Which way do you "
        "need to go? If you get stuck, don’t forget to count the blocks and check "
        "which way you need to turn each time."
    )


def title_level11():
    return "Snail maze!"


def description_level11():
    message = (
        f"Uh oh, a tricky snail maze! Can you take the van through it? "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level11(), message)


def hint_level11():
    return (
        "The warehouse is in a different place again so check which way you need to "
        "go. If you get stuck, remember that you can just add a few blocks and run "
        "the program and then add a few more and try again."
    )


def title_level12():
    return "This road is more complicated"


def description_level12():
    message = (
        f"Good work, by now you are able to solve quite complicated levels. Prove "
        f"your skills! "
        f"<br> When you are happy with your sequence, press {play_button_icon_url()}!"
    )
    return build_description(title_level12(), message)


def hint_level12():
    return (
        "You have learned a lot in these levels. Here’s a route to really test your "
        "skills. Try counting the number of straight sections and then add the "
        "curves in slowly. You can run the program before it is complete and then "
        "add to it. That can be easier than adding lots of blocks at once. Good luck!"
    )


def title_level13():
    return "Multiple routes"


def description_level13():
    message = (
        f"Often there is more than one way to get to the house. The route that needs "
        f"the fewest directions is usually best. <br> Help the van find the "
        f"shortest route to the house. <br> You can press the {play_button_icon_url()} "
        f"or <b>Play</b> buttons to start the van."
    )
    return build_description(title_level13(), message)


def hint_level13():
    return (
        "There are a few different ways of getting to the house in these levels. Can "
        "you find the shortest route? Try counting the blocks if you’re not sure. "
        "That’s the way to get the best score on these levels. "
    )


def title_level14():
    return "Can you spot the shortest route?"


def description_level14():
    message = (
        "So many options to choose from! <br> Do you know which is the shortest "
        "route to get the van to house?"
    )
    return build_description(title_level14(), message)


def hint_level14():
    return (
        "This one is a bit trickier. There are lots of different ways you could go but "
        "there is no point wasting fuel! If you take one of the longer routes, you "
        "won’t get such a good algorithm score."
    )


def title_level15():
    return "What if there is more than one delivery?"


def description_level15():
    message = (
        "Our vans often need to go to more than one house. To make the van deliver "
        "to a house use the <b>Deliver</b> block. <br> Make sure your sequence "
        "gets the van to travel the shortest route!"
    )
    return build_description(title_level15(), message)


def hint_level15():
    return (
        "On this level, there are two houses to deliver to. Which one are you going to "
        "go to first? Make sure you add a <b>Deliver</b> block for each house, you "
        "don’t want to forget anyone’s shopping!"
    )


def title_level16():
    return "This time there are even more houses"


def description_level16():
    message = (
        "Well done! You have done really well to get so far - let's take it to the "
        "next level and add another house. <br> Can you work out the shortest, "
        "most efficient route to each house?"
    )
    return build_description(title_level16(), message)


def hint_level16():
    return (
        f"Now there is another house to deliver to! Make sure you take the shortest "
        f"route and add a <b>Deliver</b> block for each house. You might find it "
        f"easier to add a small amount of code at a time and test it. You can even "
        f"use the <b>Step</b> button {step_button_icon_url()} to try each block in "
        f"your program."
    )


def title_level17():
    return "House overload!"


def description_level17():
    message = (
        "Well done, you're getting a hang of it! Can you do the same for even more "
        "houses?<br> Don't forget to use the <b>Deliver</b> block at each house."
    )
    return build_description(title_level17(), message)


def hint_level17():
    return (
        "You have four houses to deliver to now. Have you noticed that the red box "
        "turns green when you have delivered the shopping? Your programs are "
        "starting to get quite long now. Add a few blocks and then test your code. "
        "It is easier to spot errors that way. Good luck!"
    )


def title_level18():
    return "This one is quite a tangle"


def description_level18():
    message = (
        "Practise your new skills on this road by getting the van to <b>Deliver</b> "
        "to each of the houses."
    )
    return build_description(title_level18(), message)


def hint_level18():
    return (
        "You have reached the last route in this section, and it is quite complicated! "
        "Check which direction you need to turn and build up your code slowly. Maybe "
        "try to get to one house at a time and then add more code for the next one. "
        "Don’t miss a house off!"
    )


def title_level19():
    return "Repeating yourself is boring"


def description_level19():
    message = youtubeLink(
        600, 400, "https://www.youtube-nocookie.com/embed/vFGd0v3msRE", 0
    )
    message += (
        "Attach a block inside the <b>Repeat</b> block to make the van repeat "
        "that instruction. <br> This means you can use one block instead of lots "
        "of blocks to do the same thing over and over again. <br> How many times "
        "do you want the instruction repeated? Type the number into the "
        "<b>Repeat</b> block. <br> The repeated sets of blocks make a 'loop'.  "
        "<br><br> When you are ready, press <b>Play</b>!"
    )
    return build_description(title_level19(), message)


def hint_level19():
    return (
        "You have seen this route before! This time you are going to use a "
        "<b>Repeat</b> block so that you don’t need to have several "
        "<b>Move forwards</b> blocks. How many do you need?"
    )


def title_level20():
    return "Use the <b>Repeat</b> block to make your sequence shorter and simpler"


def description_level20():
    message = (
        "You drove the van down this road on Level 5. This time, use the "
        "<b>Repeat</b> block to get the van to the house. <br> This will make "
        "your sequence shorter and simpler than last time."
    )
    return build_description(title_level20(), message)


def hint_level20():
    return (
        "This is another route you have seen before. Instead of having lots of left "
        "and right blocks, can you find a pattern and put them inside a "
        "<b>Repeat</b> block? How many times do you need to repeat them?"
    )


def title_level21():
    return "Four leaf clover"


def description_level21():
    message = (
        "This path looks a bit like a four leaf clover. Can you take the driver "
        "through it?"
    )
    return build_description(title_level21(), message)


def hint_level21():
    return (
        "Here is a challenge for you! Can you find a pattern to put in the "
        "<b>Repeat</b> block? If you’re not sure, try doing your program without the "
        "repeat and then looking for the pattern. You could use a piece of paper to "
        "write the route down if that helps."
    )


def title_level22():
    return "Now things are getting quite long and complicated"


def description_level22():
    message = (
        "An algorithm (a set of instructions in a particular order) to get the van "
        "to the house might not be very simple, but it can be made shorter by "
        "using the <b>Repeat</b> blocks. <br> Are you up for this challenge?"
    )
    return build_description(title_level22(), message)


def hint_level22():
    return (
        "Instead of having lots of<b>Move forwards</b> blocks, try putting them in "
        "<b>Repeat</b> blocks when you can. This is quite a complicated route so try "
        "building up your solution slowly. If you get in a muddle, try starting "
        "again and just doing a bit of the route at a time."
    )


def title_level23():
    return "Sssssssssnake!"


def description_level23():
    message = (
        "This road seems to be winding just like a snake! Can you find a nice and "
        "simple route to get the van to the house?"
    )
    return build_description(title_level23(), message)


def hint_level23():
    return (
        "Don’t rush into this one. Try counting the straights and using a"
        "<b>Repeat</b> block for those. Then make the turn and do the same. Run the "
        "code each time."
    )


def title_level24():
    return "The road is very long and very bendy"


def description_level24():
    message = "Wow! Look at that! It won't get more complicated than this, we promise."
    return build_description(title_level24(), message)


def hint_level24():
    return (
        "This is as complicated as they get so don’t worry! Try writing down the turns "
        "to get you up to the straight bit. Can you see a pattern?"
    )


def title_level25():
    return "Waterfall level"


def description_level25():
    message = "Since you did so well with the repeat loops, have a go at this level."
    return build_description(title_level25(), message)


def hint_level25():
    return (
        "The solution to this route is quite short but you will need to find the "
        "pattern! Try solving it without loops or writing down the steps on paper. "
        "Then look for a pattern to repeat. You can do this!"
    )


def title_level26():
    return "Winter wonderland!"


def description_level26():
    message = (
        "Notice the snow! You can create new levels with different 'themes' of "
        "backgrounds and decorations in the Level Editor. But first, try getting "
        "the van to the house!"
    )
    return build_description(title_level26(), message)


def hint_level26():
    return (
        "This route is much simpler, don’t let the snow distract you! How many "
        "straights are there before the turn? How many straights after the turn?"
    )


def title_level27():
    return "Farmyard"


def description_level27():
    message = (
        "What a muddy road! Can you help Dee find her way from the barn to the "
        "house?"
    )
    return build_description(title_level27(), message)


def hint_level27():
    return (
        "Here is another type of route. This time you have a muddy road and some "
        "farmland. The pattern is not that easy to find straightaway. Try writing "
        "out the instructions without the <b>Repeat</b> block and then try to find "
        "the best pattern you can."
    )


def title_level28():
    return "The big city"


def description_level28():
    message = (
        "Can you get the van from the warehouse to the house? Don't stop at any "
        "shops on the way!"
    )
    return build_description(title_level28(), message)


def hint_level28():
    return (
        "This time you’re in a town and you need to deliver to the blue house at the "
        "end of the route. Don’t get distracted along the way. To get the best "
        "route, you need to put a <b>Repeat</b> block inside another <b>Repeat</b> "
        "block. These levels are not so simple so write down the route and then look "
        "for patterns."
    )


def title_level29():
    return "No need for numbers"


def description_level29():
    message = youtubeLink(
        600, 400, "https://www.youtube-nocookie.com/embed/EDwc80X_LQI", 0
    )
    message += (
        "Drag a block inside a <b>Repeat until</b> block to make the van repeat an "
        "instruction. <br> Attach a 'condition' so the van knows when to stop "
        "repeating the instruction. <br> Here, you want the van to repeat your "
        "instruction until it is at the destination. <br> Doing this means "
        "you don't have to work out how many times the van should repeat your "
        "instruction."
    )
    return build_description(title_level29(), message)


def hint_level29():
    return (
        "You’ve seen this route before! This time you are not counting repetitions, "
        "your loop is going to repeat the <b>Move forwards</b> block until you reach "
        "your destination."
    )


def title_level30():
    return "Can you do that again?"


def description_level30():
    message = (
        "Well done, you did it! Now have a go at using the <b>Repeat until</b> block "
        "on a road with lots of turns."
    )
    return build_description(title_level30(), message)


def hint_level30():
    return (
        "This is another route you have seen before. Last time you counted how many "
        "times your instructions were repeated. This time, your program is going to "
        "repeat your commands until you reach the destination. What do you need to "
        "repeat?"
    )


def title_level31():
    return "Practice makes perfect"


def description_level31():
    message = "Have another go to make sure you have got the hang of it."
    return build_description(title_level31(), message)


def hint_level31():
    return (
        "If you look at this route, do you notice that the shape of the road repeats? "
        "Can you put the instructions in a loop?"
    )


def title_level32():
    return "Uh oh, it's <b>Until</b> fever!"


def description_level32():
    message = "Good job! Can you help the driver reach the destination again?"
    return build_description(title_level32(), message)


def hint_level32():
    return (
        "What about this road, can you see a repeating pattern? Maybe write down the "
        "instructions without a loop to solve this route and then look for the "
        "pattern."
    )


def title_level33():
    return "Now it's time to try the <b>If</b> block"


def description_level33():
    message = youtubeLink(
        600, 400, "https://www.youtube-nocookie.com/embed/O0RXbJyYq8o", 0
    )
    message += (
        "Another way of telling the van what to do is to use the <b>If</b> block. "
        "For example, <b>If</b> the <b>road exists forwards do</b> <b>Move "
        "forwards</b>. <br> This is called an 'if statement'. <br> Try "
        "using the <b>If</b> block and the <b>Repeat</b> block together. <br> "
        "The <b>Repeat</b> block will stretch if you attach the <b>If</b> block "
        "inside it."
    )
    return build_description(title_level33(), message)


def hint_level33():
    return (
        "You could solve this with the same code you used on level 29 but the "
        "instructions for this level introduced the <b>If</b> block. To get full "
        "marks on this level, you will need to use an <b>If</b> block."
    )


def title_level34():
    return "Multiple <b>If</b>s"


def description_level34():
    message = (
        "It can be handy to use <b>If</b> to give your van choices, so you don't "
        "have to give the van new instructions at every step. <br> For "
        "example: Tell the van <b>If</b> the <b>road exists forwards do Move "
        "forwards,</b> but <b>If</b> the <b>road exists left do Turn left</b>. "
        "<br> The van will choose correctly from the <b>Move forwards</b> and "
        "<b>Turn left</b> instructions depending on the road. <br> Use an 'if "
        "statement' in a 'loop' to drive the van down this bendy road."
    )
    return build_description(title_level34(), message)


def hint_level34():
    return (
        "This route looks complicated, but you can solve it without counting blocks. "
        "You are going to use a <b>Repeat until</b> block again and <b>If</b> blocks "
        "to help the driver check the road ahead so they can decide which way to go. "
        "What are the possible directions on this route?"
    )


def title_level35():
    return "Let's put it all together!"


def description_level35():
    message = (
        "You have discovered the magic of 'if statements'. Can you make a program "
        "that uses <b>Move forwards</b>, <b>Turn left</b> and <b>Turn right</b> "
        "to get the van to the house."
    )
    return build_description(title_level35(), message)


def hint_level35():
    return (
        "For this route you need to go straight, turn left and turn right so you will "
        "need to make your <b>if</b> statement more complex. To make your code as "
        "fast as possible, think about which you need to do most. It is less "
        "efficient to ask if you need to turn left first if most of the time you "
        "want to go straight, for example."
    )


def title_level36():
    return "What else? If-else, that's what!"


def description_level36():
    message = youtubeLink(
        600, 400, "https://www.youtube-nocookie.com/embed/GUUJSRuAyU0", 0
    )
    message += (
        "You can change the <b>If</b> block to make more choices. Click on the "
        "star in the <b>If</b> block and add <b>Else if</b>. <br> This will tell "
        "the van what to do if the first <b>If</b> direction can't be done. "
        "<br> For example, tell the van to <b>Turn left</b> <b>If</b> the "
        "<b>road exists left</b>. Add <b>Else if</b> the <b>road exists right"
        "</b>, <b>Turn right</b>. <br> This uses fewer blocks and makes sure "
        "that only one step is taken in each loop. <br> This type of "
        "algorithm is called a 'general algorithm' as it can be used with most "
        "simple routes."
    )
    return build_description(title_level36(), message)


def hint_level36():
    return (
        "This route is quite different from the last one but is the solution very "
        "similar?"
    )


def title_level37():
    return "A bit longer"


def description_level37():
    message = (
        "Let's see if we can go further - this road is longer. Notice that the "
        "length of the road does not change the length of your program!"
    )
    return build_description(title_level37(), message)


def hint_level37():
    return (
        "This route is longer and a different shape again but does your last solution "
        "help? Are you noticing a pattern here?"
    )


def title_level38():
    return "Third time lucky!"


def description_level38():
    message = (
        "Well done! You've got so far. <br> Can you apply the knowledge you "
        "gained going through this part of the game to this level?"
    )
    return build_description(title_level38(), message)


def hint_level38():
    return (
        "This is a really long route. With a counted loop, your program would be quite "
        "long but is this program going to be any longer than your solution to the "
        "last level?"
    )


def title_level39():
    return "Dead ends!"


def description_level39():
    message = (
        "Can you use the 'general algorithm' here so that the van takes a "
        "shorter route? Or maybe there's a more efficient way? <br><br>Keep "
        "an eye on the fuel level - try to use as little as possible."
    )
    return build_description(title_level39(), message)


# TODO: Update when we update this level
def hint_level39():
    return (
        "Uh oh, moving around the blocks in your 'general algorithm' might not "
        "be the most efficient solution. How about creating a simple solution "
        "without 'if statements' that will help the van reach the house?"
    )


def title_level40():
    return "Adjust your previous solution"


def description_level40():
    message = (
        "Can you think of a way you could change the 'general algorithm' you have "
        "implemented earlier to make sure the van driver reaches the house having "
        "travelled the shortest route?"
    )
    return build_description(title_level40(), message)


def hint_level40():
    return (
        "With this route, only ask questions about the directions you actually need to "
        "go."
    )


def title_level41():
    return "Decision time"


def description_level41():
    message = (
        "Do you think changes to the 'general algorithm' will help the van find the "
        "shortest route? <br> Or do you have to come up with a different "
        "solution? <br> Time to make a decision..."
    )
    return build_description(title_level41(), message)


def hint_level41():
    return (
        "This is a short route and you can choose either route. Can you use an "
        "algorithm you have used before?"
    )


def title_level42():
    return "What do you think this time?"


def description_level42():
    message = (
        "Can you use the 'general algorithm' here? <br> Can it be changed so that "
        "it finds a shorter route, or will you need a new solution?"
    )
    return build_description(title_level42(), message)


def hint_level42():
    return (
        "You have been using a similar solution to lots of levels in this section, but "
        "does that help you now? Can you use a general set of instructions or do you "
        "need to write a specific solution for this one?"
    )


def title_level43():
    return "Good work! What else can you do?"


def description_level43():
    message = (
        "You should be really good at this by now. Can you manage this complicated "
        "road?"
    )
    return build_description(title_level43(), message)


def hint_level43():
    return (
        "Your general solution might not help you here. See if you need to use some "
        "counted loops for this one."
    )


def title_level44():
    return "Oh no! Traffic lights!"


def description_level44():
    message = youtubeLink(
        600, 400, "https://www.youtube-nocookie.com/embed/EDwc80X_LQI", 0
    )
    message += (
        "The light varies from red to green. <br>"
        "The van must check which colour the traffic light is when it reaches them "
        "- if it goes past a red light it will break the Highway Code."
        "<br> Here, you want the van to repeat the wait instruction while the traffic light is red. "
        "Drag a block inside a <b>Repeat while</b> block to make the van repeat an instruction. "
        "<br> Attach a 'condition' so the van knows when to repeat the instruction."
    )
    return build_description(title_level44(), message)


def hint_level44():
    return (
        "This route is very similar to one you’ve seen before but do you notice the "
        "traffic light? If the traffic light is red, you will need to wait."
    )


def title_level45():
    return "Green for go, red for wait"


def description_level45():
    message = (
        "Can you write a program so the van moves forwards on a green light but "
        "waits at a red light?"
    )
    return build_description(title_level45(), message)


def hint_level45():
    return (
        "This route is longer and there are two sets of traffic lights. Does it make "
        "any difference to your solution?"
    )


def title_level46():
    return "Well done - you've made it really far!"


def description_level46():
    message = (
        "Let's practise what you've learnt so far. <br> Don't forget to add a "
        "turn and to make the van wait at a traffic light."
    )
    return build_description(title_level46(), message)


def hint_level46():
    return (
        "In this level there is a left turn as well. Can you extend your solution to "
        "the last level to allow for that?"
    )


def title_level47():
    return "What a mess! But can you spot a route?"


def description_level47():
    message = (
        "Put your knowledge to test. Create an algorithm to lead the van to the "
        "house. <br> Don't forget to add a turn and to make the van wait at a "
        "traffic light."
    )
    return build_description(title_level47(), message)


def hint_level47():
    return (
        "This route is similar to the last one but the turn is in the other direction. "
        "Can you adapt your program?"
    )


def title_level48():
    return "Put all that hard work to the test"


def description_level48():
    message = (
        "Congratulations - you've made it really far! <br> Can you create a solution that will deliver to all of the "
        "houses in the most efficient way?"
    )
    return build_description(title_level48(), message)


def hint_level48():
    return (
        "You need to check: "
        "<ul><li> if the lights are red </li>"
        "<li> if the road exists left </li>"
        "<li> if the road exists forward </li>"
        "<li> if the road exists right </li>"
        "<li> if it is a dead end </li></ul>"
        "Make sure you put the checks in the right order. You will need to repeat this for every house and deliver."
    )


def title_level49():
    return "Amazing! Have another go!"


def description_level49():
    message = (
        "Can you change the 'general algorithm' you created before to make the van "
        "take the shortest route to the destination?"
    )
    return build_description(title_level49(), message)


def hint_level49():
    return (
        "Can you use your general algorithm to get to the house? What if you think "
        "about the order of the instructions? Of course you must obey the traffic "
        "lights but if you come to a junction, do you want to prioritise turning "
        "left, right or going straight?"
    )


def title_level50():
    return "Light maze"


def description_level50():
    message = (
        "Well this is tricky. Look at all those lights! <br> Can you find the "
        "shortest route to the destination? It would be good if the van doesn't "
        "have to wait at too many lights."
    )
    return build_description(title_level50(), message)


# TODO: Update when we update the solution and new text is provided
def hint_level50():
    return (
        "Don't worry about the algorithm you've already come up with. Take the "
        "first turn left which has fewer traffic lights. <br><br> Once your van "
        "is right under the traffic lights, make sure it waits for a green "
        "light."
    )


def title_level51():
    return "Back to basics with a twist"


def description_level51():
    message = (
        "Can you come up with a solution to this level using the limited number of blocks "
        "we provide at the start?"
    )
    return build_description(title_level51(), message)


def hint_level51():
    return (
        "In these levels, the blocks you can use are limited. Can you use the provided "
        "blocks to get to the house? If you run out blocks, check that you are "
        "taking the shortest route."
    )


def title_level52():
    return "A Bit more Tricky"


def description_level52():
    message = (
        "Well done so far! Can you find a solution to this road? You have to move forward, "
        "but you have no forward block to use. Do you know how to help the van get to "
        "the destination?"
    )
    return build_description(title_level52(), message)


def hint_level52():
    return (
        "You can’t use the shortest route here because you don’t have a "
        "<b>Move forwards</b> block. What can you do instead?"
    )


def title_level53():
    return "Choose your blocks wisely"


def description_level53():
    message = (
        "Can you find the shortest route? Use your blocks carefully and don't forget "
        "the <b>repeat</b> loop."
    )
    return build_description(title_level53(), message)


def hint_level53():
    return (
        "You must take the shortest route here otherwise you will run out of blocks. "
        "You must also use the loops to help you."
    )


def title_level54():
    return "Round and Round"


def description_level54():
    message = (
        "Can you find the shortest route? Use your blocks carefully and don't forget "
        "the <b>repeat</b> loop."
    )
    return build_description(title_level54(), message)


def hint_level54():
    return (
        "You don’t have a <b>Move forwards</b> block. Can you use the loops to get to "
        "the house? You might need to put one loop inside another loop..."
    )


def title_level55():
    return "Wonky Fish!"


def description_level55():
    message = "Use <b>repeat until</b> and the <b>if</b> statement to find your way around the Wonky Fish."
    return build_description(title_level55(), message)


def hint_level55():
    return (
        "In this level, you need to combine a <b>Repeat</b> block with an <b>If</b> "
        "block. Can you find the pattern that will get you to the house?"
    )


def title_level56():
    return "Concrete Wasteland"


def description_level56():
    message = (
        "Use <b>repeat until</b> and the <b>if</b> statement to find your way around "
        "the Concrete Wasteland"
    )
    return build_description(title_level56(), message)


def hint_level56():
    return (
        "Think about which turns you need to make to get to the destination. Which way "
        "do you go most often?"
    )


def title_level57():
    return "This is <b>not...</b> the same"


def description_level57():
    message = (
        "Like <b>repeat until</b>, <b>repeat while</b> is the opposite. Here, you want "
        "the van to repeat your instructions while it is not at the destination.<br>"
        "Doing this means you don't have to work out how many times the van should "
        "repeat your instructions."
    )
    return build_description(title_level57(), message)


def hint_level57():
    return (
        "This time you have a <b>Repeat while</b> block. This is different from the "
        "<b>Repeat until</b> you have used before. What needs to change in your code?"
    )


def title_level58():
    return "Snow snake"


def description_level58():
    message = (
        "Combining what you have just learnt using <b>repeat while</b> with the repeat "
        "loop, can you find your way around the snow snake?"
    )
    return build_description(title_level58(), message)


def hint_level58():
    return (
        "In this example, you only have four counted loop blocks. How can you use them "
        "to get to the destination?"
    )


def title_level59():
    return "Tricky turnaround"


def description_level59():
    message = "Use your blocks carefully not forgetting the <b>turnaround</b>."
    return build_description(title_level59(), message)


def hint_level59():
    return (
        "In this level, there is no <b>Move forwards</b> block. How can you use the "
        "loop, <b>Turn left</b> and <b>Turn around</b> blocks to get to the house? "
        "Try using the blocks without the loop and see where that takes you. Would "
        "repeating that help?"
    )


def title_level60():
    return "Right around the block"


def description_level60():
    message = "Can you find your way around this puzzle?"
    return build_description(title_level60(), message)


def hint_level60():
    return (
        "Can you see a pattern in this route? Use that to create a loop. When you have "
        "done that, what is left to do?"
    )


def title_level61():
    return "Can you create the 'Wiggle' procedure?"


def description_level61():
    message = (
        "Procedures are groups of instructions that can be executed multiple times "
        "without being rewritten. For example, if you want to instruct the van to "
        "follow a repeated pattern in the road, you can create a specific procedure. "
        "To create a procedure, simply choose the correct blocks and put them in the "
        "right order inside the <b>Define do</b> block. Once you have done that, give "
        "it a name eg wiggle.<br>Now you're ready! Attach the <b>Call</b> block where "
        "you want your 'wiggle' procedure to be executed. Don't forget to put the name in it!"
    )
    return build_description(title_level61(), message)


def hint_level61():
    return (
        "The instructions asked you to create a procedure called ‘wiggle’. This "
        "procedure should make the moves to manage the bends in the road. You will "
        "need to call this procedure from the main program."
    )


def title_level62():
    return "Lots of Traffic Lights!"


def description_level62():
    message = "Create a procedure which tells the van to wait until the traffic lights are green."
    return build_description(title_level62(), message)


def hint_level62():
    return (
        "Can you create a procedure that just deals with traffic lights? Call it "
        "‘lights’ and call it from the main program whenever you need it. Don’t "
        "worry if you don’t use the <b>Repeat block</b> in the program."
    )


def title_level63():
    return "Wiggle Wiggle"


def description_level63():
    message = (
        "Can you find the repeating pattern here and create a new 'wiggle' procedure? "
        "And do the Wiggle Wiggle!"
    )
    return build_description(title_level63(), message)


def hint_level63():
    return (
        "In this level, you should notice a repeated turn in the road. Can you create "
        "a ‘wiggle’ procedure and call it whenever you need it from the main "
        "program? If you can’t find the repeated pattern, try writing the route out "
        "in full and looking at your instructions."
    )


def title_level64():
    return "Muddy Patterns with Phil"


def description_level64():
    message = (
        "Can you spot a pattern here? Create several procedures, it can save time when "
        "writing a program. Don't forget to clearly name your procedures and then call them."
    )
    return build_description(title_level64(), message)


def hint_level64():
    return (
        "Try to create two procedures to help you here. Then call them as you need "
        "them from the main procedure. You could use one procedure to deal with each "
        "of the turn patterns in the route."
    )


def title_level65():
    return "Complicated roads"


def description_level65():
    message = (
        "This road might be a bit more complicated, but the procedures you could come up "
        "with are quite simple. Have a go and find out yourself!"
    )
    return build_description(title_level65(), message)


def hint_level65():
    return (
        "This route involves using the main program to deal with the main route and "
        "using procedures to deal with the turns in the road. You could try creating "
        "the route without procedures and then looking for a pattern to move into a "
        "procedure."
    )


def title_level66():
    return "Dee's snowy walk"


def description_level66():
    message = "Did you know procedures can call other procedures?"
    return build_description(title_level66(), message)


def hint_level66():
    return (
        "In this example, you can write a procedure that calls another procedure. Look "
        "for certain actions that are repeated and try to put them in procedures."
    )


def title_level67():
    return "Crazy Farm"


def description_level67():
    message = "This one will really test what you have learnt."
    return build_description(title_level67(), message)


def hint_level67():
    return (
        "In this level, try to find several patterns and put them in procedures. How "
        "about starting with a wiggle?"
    )


def title_level68():
    return "T - time"


def description_level68():
    message = "Can you find the shortest route?"
    return build_description(title_level68(), message)


def hint_level68():
    return (
        "You have limited blocks again here. How can you get to the house without a "
        "counted loop?"
    )


def title_level69():
    return "Duck pond dodge"


def description_level69():
    message = "Can you find the shortest route?"
    return build_description(title_level69(), message)


def hint_level69():
    return (
        "In this level, you need a loop inside another loop. When you are able to go "
        "forwards, you should keep going. When you can no longer go forwards, you "
        "need to turn and then try to go forwards again.<br><br>Hint: The traffic "
        "lights are a bit of a distraction and you don’t need to worry about them!"
    )


def title_level70():
    return "Winter wonderland"


def description_level70():
    message = "Can you find the shortest route?"
    return build_description(title_level70(), message)


def hint_level70():
    return (
        "You don’t have any <b>Move forwards</b> blocks here. How can you use a loop "
        "and the turn blocks to get to the house?"
    )


def title_level71():
    return "Frozen challenge"


def description_level71():
    message = "Can you find the shortest route?"
    return build_description(title_level71(), message)


def hint_level71():
    return (
        "In this level, most of the time you want to <b>Move forwards</b> and then "
        "<b>turn right</b>. However, there is an occasion when you don’t want to do "
        "that."
    )


def title_level72():
    return "Can Wes Find his lunch?"


def description_level72():
    message = "Can you find the shortest route?"
    return build_description(title_level72(), message)


def hint_level72():
    return (
        "In this level, you want to <b>Move forwards</b> most of the time. When do you "
        "need to turn?"
    )


def title_level73():
    return "Traffic light freeze up!"


def description_level73():
    message = "Can you find the shortest algorithm?"
    return build_description(title_level73(), message)


def hint_level73():
    return (
        "You don’t have the blocks to check if you are at the destination. What can "
        "you use instead?  Don’t forget to deal with the traffic lights!"
    )


def title_level74():
    return "Pandemonium"


def description_level74():
    message = "Can you find the shortest route?"
    return build_description(title_level74(), message)


def hint_level74():
    return (
        "You only have one <b>Repeat</b> block here so make the most of it! You can "
        "use procedures to repeat code by just calling the procedure again."
    )


def title_level75():
    return "Kirsty's maze time"


def description_level75():
    message = "Can you find the shortest route?"
    return build_description(title_level75(), message)


def hint_level75():
    return (
        "You need to look for the shortest route here and then use loops inside loops "
        "to get you there. Good luck!"
    )


def title_level76():
    return "Cannot turn left!"


def description_level76():
    message = "Can you find the shortest route?"
    return build_description(title_level76(), message)


def hint_level76():
    return (
        "This van driver is going to have to take quite a circuitous route to make "
        "their delivery. You don’t have any left turns!"
    )


def title_level77():
    return "G Force"


def description_level77():
    message = "Can you get the van to the house?"
    return build_description(title_level77(), message)


def hint_level77():
    return (
        "You don’t have any loops here but you can make a subroutine. You know that a "
        "subroutine can call another subroutine but did you know that a subroutine "
        "can call itself?"
    )


def title_level78():
    return "Wandering Phil"


def description_level78():
    message = "Can you get Phil to the house?"
    return build_description(title_level78(), message)


def hint_level78():
    return (
        "You really need to put together everything you have learned for this tricky "
        "level. You need to combine checking for being at the destination with "
        "checking for a dead end and which roads are available. Good luck!"
    )


def title_level79():
    return "Muddy Mayhem"


def description_level79():
    message = "Can you find the shortest route?"
    return build_description(title_level79(), message)


def hint_level79():
    return (
        "This is the last of the levels in this section and you will have to put "
        "together everything you have learned! You don’t have any <b>Repeat</b> "
        "blocks so you need to use a procedure that calls itself and also checks "
        "which roads are available."
    )


# --- Introduction to Python levels ---

INTRO_HINT = (
    "If you're stuck, you can learn more about the use of Blockly and Python on "
    "<a href='https://docs.codeforlife.education/rapid-router' target='_blank'>our documentation site</a>."
    "<br /><br />To learn more about Python in general, check this "
    "<a href='https://wiki.python.org/moin/BeginnersGuide' target='_blank'>Beginner's Guide to Python</a>."
)


def title_level80():
    return "Here's Python!"


def description_level80():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Can you tell which Python statement matches which block?"
    )
    return build_description(title_level80(), message)


def hint_level80():
    return (
        "This route is quite simple and you have seen it before. Just drag the blocks "
        "as you normally would but notice what is happening in the Python code…"
    )


def title_level81():
    return "Matching Blockly"


def description_level81():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Can you tell which Python statement matches which block?"
    )
    return build_description(title_level81(), message)


# TODO: Update once we've renamed v to my_van
def hint_level81():
    return INTRO_HINT


def title_level82():
    return "Don't forget to find the shortest route"


def description_level82():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Can you tell which Python statement matches which block?"
    )
    return build_description(title_level82(), message)


def hint_level82():
    return (
        "You can’t edit the Python program yet but have a look at it as you add new "
        "blocks to your code. Notice that each command to the van ends with round "
        "brackets (). These are important, they tell Python that you are giving a "
        "command."
    )


def title_level83():
    return "Repeating yourself in Python looks different"


def description_level83():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding a <b>repeat</b> block and watch what happens in Python."
    )
    return build_description(title_level83(), message)


# TODO: Update once we change the solution for this level
def hint_level83():
    return INTRO_HINT


def title_level84():
    return "Repeat and watch"


def description_level84():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding a <b>repeat</b> block and watch what happens in Python."
    )
    return build_description(title_level84(), message)


# TODO: Update once we change the solution for this level
def hint_level84():
    return INTRO_HINT


def title_level85():
    return "Looks easy but use repeat until and see what happens?"


def description_level85():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding a <b>repeat</b> until block and watch what "
        "happens in Python."
    )
    return build_description(title_level85(), message)


# TODO: For levels 85 - 109: the new hint text has not yet been provided.
def hint_level85():
    return INTRO_HINT


def title_level86():
    return "See what the if blocks looks like in Python"


def description_level86():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding an <b>if</b> block and watch what happens in Python."
    )
    return build_description(title_level86(), message)


def hint_level86():
    return INTRO_HINT


def title_level87():
    return "Don't forget to use else if"


def description_level87():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding an <b>if</b> block and watch what happens in "
        "Python particularly with <b>else if</b> and <b>else</b> statements."
    )
    return build_description(title_level87(), message)


def hint_level87():
    return INTRO_HINT


def title_level88():
    return "See what happens when you add Traffic lights"


def description_level88():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding an <b>if</b> block and watch what happens in "
        "Python particularly with <b>else if</b> and <b>else</b> statements."
    )
    return build_description(title_level88(), message)


def hint_level88():
    return INTRO_HINT


def title_level89():
    return "Watch carefully as you have another go"


def description_level89():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "programming language. Try adding an <b>if</b> block and watch what happens in "
        "Python particularly with <b>else if</b> and <b>else</b> statements."
    )
    return build_description(title_level89(), message)


def hint_level89():
    return INTRO_HINT


def title_level90():
    return "Have a go at procedures - what do they look like in Python?"


def description_level90():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "language. Try adding a procedure and watch what happens in Python."
    )
    return build_description(title_level90(), message)


def hint_level90():
    return (
        "Don't forget to name your procedure and see what happens in Python. "
        + INTRO_HINT
    )


def title_level91():
    return "Put it all together"


def description_level91():
    message = (
        "As you create your program using Blockly see what it looks like in the Python "
        "language. Try adding a procedure and watch what happens in Python."
    )
    return build_description(title_level91(), message)


def hint_level91():
    return (
        "Don't forget to name your procedure and see what happens in Python. "
        + INTRO_HINT
    )


# --- Start of Python levels ---

PYTHON_HINT = (
    "<br /><br />Check our documentation site, to see "
    "<a href='https://docs.codeforlife.education/rapid-router/python-commands' target='_blank'>the full list of commands</a>."
    "<br /><br />To learn more about Python in general, check this "
    "<a href='https://wiki.python.org/moin/BeginnersGuide' target='_blank'>Beginner's Guide to Python</a>."
)


def title_level92():
    return "Start with the basics, <b>forward</b>, <b>left</b> and <b>right</b>"


def description_level92():
    message = (
        "Now you are coding in Python! This is what real developers do!! To start you off, "
        "the van object has been created for you already. Under this you need to add the "
        "correct Python statements to instruct the van to drive to the destination."
    )
    return build_description(title_level92(), message)


def hint_level92():
    return (
        "Try using the following commands:<br><pre>my_van.move_forwards()<br>my_van.turn_left()"
        "<br>my_van.turn_right()</pre>" + PYTHON_HINT
    )


def title_level93():
    return "Keep it simple"


def description_level93():
    message = (
        "Try this road. Under the van object you need to add the correct Python statements "
        "to instruct the van to drive to the destination."
    )
    return build_description(title_level93(), message)


def hint_level93():
    return (
        """Try using the following commands:
<pre>my_van.move_forwards()
my_van.turn_left()
my_van.turn_right()</pre>"""
        + PYTHON_HINT
    )


def title_level94():
    return "Take the shortest route"


def description_level94():
    message = (
        "You're getting good at this! Can you drive the van along this road using the "
        "correct Python statements."
    )
    return build_description(title_level94(), message)


def hint_level94():
    return (
        """Try using the following commands:
<pre>my_van.move_forwards()
my_van.turn_left()
my_van.turn_right()</pre>"""
        + PYTHON_HINT
    )


def title_level95():
    return "Count and repeat"


def description_level95():
    message = (
        "Now try to use a <b>repeat</b> loop to solve this level. Look back at level 83 "
        "to see what this could look like in Python."
    )
    return build_description(title_level95(), message)


def hint_level95():
    return (
        """To repeat some statements a set number of times you can use something like the following:
<pre>for count in range(3):
    my_van.turn left
    print count</pre>
The print statement will output the value of count to the console."""
        + PYTHON_HINT
    )


def title_level96():
    return "Count and repeat is easy"


def description_level96():
    message = (
        "Now try to use a <b>repeat loop</b> to solve this level. Look back at level 83 "
        "to see what this could look like in Python. This time you could use 2 loops, "
        "1 for each straight piece of road."
    )
    return build_description(title_level96(), message)


def hint_level96():
    return (
        """To repeat some statements a set number of times you can use something like the following:
<pre>for count in range(3):
    my_van.turn left
    print count</pre>
The print statement will output the value of count to the console."""
        + PYTHON_HINT
    )


def title_level97():
    return "Loop the loop"


def description_level97():
    message = (
        "Now try to use a loop within a loop, known as a 'nested loop'. Look back at level "
        "84 to see what this could look like in Python."
    )
    return build_description(title_level97(), message)


def hint_level97():
    return (
        """To repeat within a repeats a set number of times you can use something like the following:
<pre>for i in range(3):
    for j in range(5):
        my_van.turn left
        print count</pre>
The print statement will output the value of count to the console."""
        + PYTHON_HINT
    )


def title_level98():
    return "Repeat and check"


def description_level98():
    message = (
        "Try to solve this level by repeatedly moving until the van is at the destination. "
        "Also, check whether the van can move forward or else must turn left. Now try and "
        "write the Python code. Look back at level 86 to give you an idea of what this "
        "could look like."
    )
    return build_description(title_level98(), message)


def hint_level98():
    return (
        """To repeat while a condition is met you can use something like the following:
<pre>while not my_van.at_destination():
    my_van.move_forwards()</pre>
To check whether a condition is met you can use something like the following:
<pre>if my_van.is_road_forward():
    my_van.move_forwards()</pre>
You may also need to use the <b>else</b> statement."""
        + PYTHON_HINT
    )


def title_level99():
    return "Find a general solution"


def description_level99():
    message = (
        "Now try using what you have just learnt to solve this level. You could also try "
        "using the <b>if</b>, <b>elif</b> and <b>else</b> statements. Look back at level "
        "86 to give you an idea of what this could look like."
    )
    return build_description(title_level99(), message)


def hint_level99():
    return (
        """To repeat while a condition is met you can use something like the following:
<pre>while not my_van.at_destination():
    my_van.move_forwards()</pre>
To check whether a condition is met you can use something like the following:
<pre>if my_van.is_road_forward():
    my_van.move_forwards()</pre>
You may also need to use the <b>elif</b> and <b>else</b> statements."""
        + PYTHON_HINT
    )


def title_level100():
    return "Watch out for the dead end!"


def description_level100():
    message = (
        "Practice your new Python skills on this road to get the van to the destination. "
        "Look back at level 88 for a dead end check."
    )
    return build_description(title_level100(), message)


def hint_level100():
    return (
        "Try using<br><pre>if my_van.at_dead_end():</pre><br>to check if the van is at a dead end."
        + PYTHON_HINT
    )


def title_level101():
    return "Function or Junction?"


def description_level101():
    message = (
        "Try defining your own procedure to solve this level. In Python procedures are "
        "generally called functions. Look back at level 90 for an example of how to "
        "define a function in Python."
    )
    return build_description(title_level101(), message)


def hint_level101():
    return (
        """To define a function in Python you could do something like:
<pre>def my_function():
    print 'test'</pre>
To call a defined function you could do something like:
<pre>my_function()</pre>
Remember, you must define a function before you call it."""
        + PYTHON_HINT
    )


def title_level102():
    return "Watch for the patterns"


def description_level102():
    message = (
        "For this level try defining more than one function. Try to look for a repeating "
        "pattern to simplify your program."
    )
    return build_description(title_level102(), message)


def hint_level102():
    return (
        """To define a function in Python you could do something like:
<pre>def my_function():
    print 'test'</pre>
To call a defined function you could do something like:
<pre>my_function()</pre>"""
        + PYTHON_HINT
    )


def title_level103():
    return "Patterns within patterns"


def description_level103():
    message = (
        "For this level try to define 2 or more functions where inside one function you "
        "call another function."
    )
    return build_description(title_level103(), message)


def hint_level103():
    return (
        """To define a function that calls another function you could do something like:
<pre>def my_function():
    print 'test'

def my_other_function():
    for i in range(3):
        my_function()

my_other_function()</pre>"""
        + PYTHON_HINT
    )


def title_level104():
    return "Can you see the repeating pattern?"


def description_level104():
    message = (
        "For this level try to define 2 or more functions where inside one function you "
        "call another function."
    )
    return build_description(title_level104(), message)


def hint_level104():
    return (
        """To define a function that calls another function you could do something like:
<pre>def my_function():
    print 'test'

def my_other_function():
    for i in range(3):
        my_function()

my_other_function()</pre>"""
        + PYTHON_HINT
    )


def title_level105():
    return "Find the shortest route"


def description_level105():
    message = (
        "For this level try to implement a general algorithm. Keep the van going until it "
        "arrives at the destination, checking for traffic lights and junctions."
    )
    return build_description(title_level105(), message)


def hint_level105():
    return (
        "For this you will have to use a combination of the <b>while</b> and <b>if</b> statements."
        + PYTHON_HINT
    )


def title_level106():
    return "Spiral and add"


def description_level106():
    message = (
        "For this level the van needs to travel in a spiral. The number of grid squares the "
        "van has to move keeps increasing by 1 on each turn. To do this you can have a loop "
        "that makes use of a variable to track the length of the road you need to travel "
        "after each turn."
    )
    return build_description(title_level106(), message)


def hint_level106():
    return (
        """To use a variable to store the number of grid squares the van has to move you can do something like the following:
<pre>n = 1
while not my_van.at_destination():
    print n
    n += 1</pre>
Variables can be used in place of constants when calling functions. For example to repeat something n times you can do something like the following:
<pre>for count in range(n):</pre>"""
        + PYTHON_HINT
    )


def title_level107():
    return "Spiral and double"


def description_level107():
    message = (
        "For this level try something similar to what you have just learnt. This time "
        "the straight sections of road are doubling in length after each turn."
    )
    return build_description(title_level107(), message)


def hint_level107():
    return (
        "To double the value of a variable you can do something like the following:<br><pre>n *= 2</pre>"
        + PYTHON_HINT
    )


def title_level108():
    return "Think less"


def description_level108():
    message = "This time the straight sections of road decrease in length by 2 after each turn."
    return build_description(title_level108(), message)


def hint_level108():
    return (
        "To decrease the value of a variable by an amount you can do something like the "
        "following:<br><pre>n -= 5</pre>" + PYTHON_HINT
    )


def title_level109():
    return "Final challenge!"


def description_level109():
    message = (
        "For the last challenge, the road straight line sections of road start off increasing "
        "by 1 after each turn and then switch to dividing by 2 with a twist!"
    )
    return build_description(title_level109(), message)


def hint_level109():
    return (
        "To halve the value of a variable you can do something like the following:<br><pre>n /= 2</pre>"
        + PYTHON_HINT
    )


def get_episode_title(episode_id):
    episode_titles = {
        1: "Getting Started",
        2: "Shortest Route",
        3: "Loops and Repetitions",
        4: "Loops with Conditions",
        5: "If... Only",
        6: "Traffic Lights",
        7: "Limited Blocks",
        8: "Procedures",
        9: "Blockly Brain Teasers",
        10: "Introduction to Python",
        11: "Python",
    }

    return episode_titles[episode_id]



from django.db import migrations


def set_python_view(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    levels = [63, 64, 65, 66, 67, 68, 69, 93, 94, 95, 96, 97]
    for level_id in levels:
        level = Level.objects.get(pk=level_id)
        level.pythonEnabled = False
        level.pythonViewEnabled = True
        level.blocklyEnabled = True
        level.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0030_merge")]

    operations = [migrations.RunPython(set_python_view)]



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open("game/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="rapid-router",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django==3.2.13",
        "django-csp==3.7",
        "django-js-reverse==0.9.1",
        "django-pipeline==2.0.8",
        "djangorestframework==3.13.1",
        "more-itertools==8.6.0",
        "pyhamcrest==2.0.2",
        "libsass==0.21.0",
        "future==0.18.2",
        "cfl-common",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django",
    ],
    version=version,
    zip_safe=False,
)



from __future__ import unicode_literals

from builtins import map, range
from datetime import datetime, timedelta

from common.models import Class, Teacher, Student
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import (
    create_school_student_directly,
    create_independent_student_directly,
)
from common.tests.utils.teacher import signup_teacher_directly
from django.test import Client, TestCase
from django.urls import reverse

from game.models import Attempt, Level, Episode
from game.views.scoreboard import StudentRow, scoreboard_data, Headers, StudentInTrouble
from game.views.scoreboard_csv import scoreboard_csv, Headers as CSVHeaders


class ScoreboardTestCase(TestCase):
    def test_teacher_multiple_students_multiple_levels(self):
        episode_ids = [1]
        episode1 = Episode.objects.get(id=1)
        level_ids = [f"{x}" for x in range(1, len(episode1.levels) + 1)]
        level1 = Level.objects.get(name="1")
        level2 = Level.objects.get(name="2")

        clas, student, student2 = set_up_data()

        create_attempt(student, level1, 10)
        create_attempt(student2, level1, 2)
        create_attempt(student2, level2, 16)

        all_levels = [level1, level2]

        attempts_per_student = {
            student: Attempt.objects.filter(
                level__in=all_levels, student=student, is_best_attempt=True
            ).select_related("level"),
            student2: Attempt.objects.filter(
                level__in=all_levels, student=student2, is_best_attempt=True
            ).select_related("level"),
        }

        student_data, headers, level_headers, levels_sorted = scoreboard_data(
            episode_ids, attempts_per_student
        )

        assert headers == Headers
        assert level_headers == [f"L{id}" for id in level_ids]

        assert len(student_data) == 2

        student_row = student_data[0]
        assert student_row.class_field.name == clas.name
        assert student_row.name == student.user.user.first_name
        assert student_row.total_score == 10
        assert student_row.total_time == timedelta(0)
        assert student_row.level_scores[1]["score"] == 10
        assert student_row.completed == 1
        assert student_row.percentage_complete == 50.0

        student_row = student_data[1]
        assert student_row.class_field.name == clas.name
        assert student_row.name == student2.user.user.first_name
        assert student_row.total_score == 18
        assert student_row.total_time == timedelta(0)
        assert student_row.level_scores[1]["score"] == 2
        assert student_row.level_scores[2]["score"] == 16
        assert student_row.completed == 1
        assert student_row.percentage_complete == 45.0

    def test_scoreboard_loads(self):
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, name, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        url = reverse("scoreboard")
        c = Client()
        c.login(username=email, password=password)

        # test scoreboard page loads properly
        response = c.get(url)
        assert response.status_code == 200

        # test scoreboard shows all episodes if no episodes are manually selected
        data = {"classes": [klass.id], "view": [""]}

        response = c.post(url, data)
        assert response.status_code == 200
        assert len(response.context["level_headers"]) == 109

    def test_student_can_see_classes(self):
        """A student should be able to see the classes they are in"""
        mr_teacher = Teacher.objects.factory(
            "Normal", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        klass, name1, _ = create_class_directly(
            mr_teacher.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            mr_teacher.user.user.email, class_name="Class 2"
        )
        student = Student.objects.schoolFactory(klass, "some student", "secret")

        c = Client()
        c.force_login(student.user.user)

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]
        assert name1 in choices_in_form
        assert name2 not in choices_in_form
        assert len(choices_in_form) == 1

    def test_admin_teacher_can_see_all_classes(self):
        """An admin should be able to see all classes, not just the ones they teach"""
        normal_teacher = Teacher.objects.factory(
            "Normal", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        admin_teacher = Teacher.objects.factory(
            "Admin", "Admin", "admin@school.edu", "secretpa$sword2"
        )

        admin_teacher.is_admin = True
        admin_teacher.save()

        _, name1, _ = create_class_directly(
            admin_teacher.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            admin_teacher.user.user.email, class_name="Class 2"
        )
        _, name3, _ = create_class_directly(
            normal_teacher.user.user.email, class_name="Class 3"
        )

        c = Client()
        c.login(username=admin_teacher.user.user.email, password="secretpa$sword2")

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name1 in choices_in_form
        assert name2 in choices_in_form
        assert name3 in choices_in_form

    def test_non_admin_teacher_can_only_see_their_own_classes(self):
        """A teacher who is not an admin should only be able to see their classes, not ones taught by others"""
        teacher1 = Teacher.objects.factory(
            "First", "Teacher", "normal@school.edu", "secretpa$sword"
        )
        teacher2 = Teacher.objects.factory(
            "Second", "Teacher", "admin@school.edu", "secretpa$sword2"
        )

        _, name1, _ = create_class_directly(
            teacher2.user.user.email, class_name="Class 1"
        )
        _, name2, _ = create_class_directly(
            teacher2.user.user.email, class_name="Class 2"
        )
        _, name3, _ = create_class_directly(
            teacher1.user.user.email, class_name="Class 3"
        )

        c = Client()
        # First teacher logs in.  Should see only Class 3
        c.login(username="normal@school.edu", password="secretpa$sword")

        url = reverse("scoreboard")
        response = c.get(url)

        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name3 in choices_in_form
        assert name1 not in choices_in_form
        assert name2 not in choices_in_form
        assert len(choices_in_form) == 1

        # Other teacher logs in.  Should see Class 1 and Class 2
        c.login(username="admin@school.edu", password="secretpa$sword2")

        response = c.get(url)
        choices_in_form = [
            v for (k, v) in response.context["form"]["classes"].field.choices
        ]

        assert name3 not in choices_in_form
        assert name1 in choices_in_form
        assert name2 in choices_in_form
        assert len(choices_in_form) == 2

    def test_independent_student_cannot_see_scoreboard(self):
        username, password, _ = create_independent_student_directly()

        c = Client()
        c.login(username=username, password=password)

        url = reverse("scoreboard")
        response = c.get(url)

        assert "Scoreboard is only visible to school students and teachers" in str(
            response.content
        )


class ScoreboardCsvTestCase(TestCase):
    def test_scoreboard_csv(self):
        levels = Level.objects.sorted_levels()[0:2]

        student_rows = [None, None]
        students = [None, None]
        student_rows[0], students[0] = self.student_row("secrète")
        student_rows[1], students[1] = self.student_row()

        improvement_data = []
        for student in students:
            stud = StudentInTrouble(student=student, areas="Getting started")
            improvement_data.append(stud)

        response = scoreboard_csv(student_rows, levels, improvement_data)

        actual_header, actual_rows = self.actual_data(response.content.decode("utf-8"))

        expected_header = self.expected_header(levels)
        assert actual_header == expected_header

        assert len(actual_rows) == len(student_rows)

        # check first row
        (
            class_name,
            name,
            completed_levels,
            total_time,
            total_scores,
            l1,
            l2,
            improvement,
        ) = actual_rows[0].split(",")
        assert student_rows[0].class_field.name == class_name
        assert student_rows[0].name == name
        assert student_rows[0].level_scores[0]["score"] == int(l1)
        assert student_rows[0].level_scores[1]["score"] == int(l2)
        assert improvement == "Getting started"

        # check last row
        last = len(actual_rows) - 1
        (
            class_name,
            name,
            completed_levels,
            total_time,
            total_scores,
            l1,
            l2,
            improvement,
        ) = actual_rows[last].split(",")
        assert student_rows[last].class_field.name == class_name
        assert student_rows[last].name == name
        assert str(student_rows[last].total_time) == total_time

    def student_row(self, class_name=None):
        email, password = signup_teacher_directly()
        _, class_name, access_code = create_class_directly(email, class_name)
        _, _, student = create_school_student_directly(access_code)

        total_time = timedelta(seconds=30)
        scores = [x for x in range(20)]
        total_score = sum(scores)

        level_scores = {}
        for i in range(2):
            level_scores[i] = {}
            level_scores[i]["score"] = 5
            level_scores[i]["full_score"] = 20
            level_scores[i]["is_low_attempt"] = True

        row = StudentRow(
            student=student,
            class_field=Class(name="MyClass"),
            total_time=total_time,
            total_score=total_score,
            level_scores=level_scores,
            completed=2,
            percentage_complete=45,
        )

        return row, student

    def expected_header(self, levels):
        level_strings = list(map(str, levels))
        all_header_strings = CSVHeaders + level_strings + ["Areas for improvement"]
        joined = ",".join(all_header_strings)
        return joined

    def actual_data(self, content):
        split = content.strip().split("\r\n")
        header = split[0]
        rows = split[1:]
        return header, rows


def create_attempt(student, level, score):
    attempt = Attempt.objects.create(
        finish_time=datetime.fromtimestamp(1435305072),
        level=level,
        student=student,
        score=score,
        is_best_attempt=True,
    )
    attempt.start_time = datetime.fromtimestamp(1435305072)
    attempt.save()


def ids_of_levels_named(names):
    levels = Level.objects.filter(name__in=names)
    return levels


def set_up_data(classmates_data_viewable=False):
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    if classmates_data_viewable:
        clas.classmates_data_viewable = True
        clas.save()
    _, _, student = create_school_student_directly(access_code)
    _, _, student2 = create_school_student_directly(access_code)
    create_random_school_data()
    return clas, student, student2


def create_random_school_data():
    email, password = signup_teacher_directly()
    clas, class_name, access_code = create_class_directly(email)
    create_school_student_directly(access_code)
    create_school_student_directly(access_code)



from game.end_to_end_tests.base_game_test import BaseGameTest


class TestLevelWin(BaseGameTest):
    def test_deliver_everywhere(self):
        self.deliver_everywhere_test(level=16)

    def test_try_again_if_not_full_score(self):
        page = self.try_again_if_not_full_score_test(
            level=19, workspace_file="complete_level_not_with_full_score"
        )
        self.complete_and_check_level(19, page)



from __future__ import absolute_import
from __future__ import division

from builtins import map
from builtins import next
from builtins import object
from datetime import timedelta

from common.models import Class, Teacher, Student
from django.http import Http404
from django.shortcuts import render

import game.messages as messages
import game.permissions as permissions
from game.forms import ScoreboardForm
from game.models import Level, Attempt, sort_levels, Episode
from game.views.scoreboard_csv import scoreboard_csv
from . import level_selection
from .helper import renderError

Headers = ["Class", "Student", "Completed", "Total time"]
TotalPointsHeader = "Total points"


class StudentRow:
    def __init__(self, *args, **kwargs):
        student = kwargs.get("student")
        self.class_field = student.class_field
        self.name = student.user.user.first_name
        self.id = student.id
        self.total_time = kwargs.get("total_time", timedelta(0))
        self.total_score = kwargs.get("total_score", 0)
        self.level_scores = kwargs.get("level_scores", {})
        self.completed = kwargs.get("completed", 0)
        self.percentage_complete = kwargs.get("percentage_complete", 0)


# Returns row of student object with values and scores of each selected level
def student_row(levels_sorted, student, best_attempts):
    threshold = 0.5

    num_all = num_finished = num_attempted = num_started = 0
    total_score = 0
    total_possible_score = 0
    level_scores = {}
    times = []

    for level in levels_sorted:
        level_scores[level.id] = {}
        level_scores[level.id]["score"] = ""

    if best_attempts:
        attempts_dict = {
            best_attempt.level.id: best_attempt for best_attempt in best_attempts
        }
        for level in levels_sorted:

            attempt = attempts_dict.get(level.id)
            if attempt:
                num_all += 1
                max_score = 10 if attempt.level.disable_route_score else 20
                if attempt.score:
                    if attempt.score / max_score >= threshold:
                        num_finished += 1
                    else:
                        num_attempted += 1
                else:
                    num_started += 1

                total_score += attempt.score if attempt.score is not None else 0
                total_possible_score += max_score

                elapsed_time = attempt.elapsed_time()
                times.append(chop_miliseconds(elapsed_time))
                # '-' is used to show that the student has started the level but has not submitted any attempts

                level_scores[level.id]["score"] = (
                    int(attempt.score) if attempt.score is not None else "-"
                )
                level_scores[level.id]["full_score"] = attempt.score == max_score
                level_scores[level.id]["is_low_attempt"] = (
                    attempt.score == 0 or max_score / attempt.score < threshold
                )
            else:
                times.append(timedelta(0))

    total_time = sum(times, timedelta())
    percentage_complete = (
        total_score / total_possible_score * 100 if total_possible_score > 0 else 0
    )

    row = StudentRow(
        student=student,
        total_time=total_time,
        total_score=int(total_score),
        level_scores=level_scores,
        completed=num_finished,
        percentage_complete=percentage_complete,
    )
    return row


def to_name(level):
    return f"L{level.name}"


def scoreboard_data(episode_ids, attempts_per_students):
    # Show the total score, total time and score of each level
    levels_sorted = []
    for episode_id in episode_ids:
        episode = Episode.objects.get(id=episode_id)
        levels_sorted += episode.levels

    level_headers = list(map(to_name, levels_sorted))
    student_data = [
        student_row(levels_sorted, student, best_attempts)
        for student, best_attempts in attempts_per_students.items()
    ]

    return student_data, Headers, level_headers, levels_sorted


class StudentInTrouble:
    def __init__(self, *args, **kwargs):
        student = kwargs.get("student")
        self.class_field = student.class_field
        self.name = student.user.user.first_name
        self.areas = kwargs.get("areas")


def _check_attempts(best_attempts):
    threshold = 0.5

    # episode ids with low attempts (below 50%)
    low_episode_ids = set()

    for episode_id in range(1, 12):
        total_score = 0
        total_possible_score = 0
        # Get the best attempts for the specific Episode
        attempts = [
            best_attempt
            for best_attempt in best_attempts
            if best_attempt.level.episode.id == episode_id
        ]
        for attempt in attempts:
            max_score = 10 if attempt.level.disable_route_score else 20

            total_score += attempt.score if attempt.score is not None else 0
            total_possible_score += max_score

            is_low_attempt = attempt.score == 0 or max_score / attempt.score < threshold
            if is_low_attempt:
                low_episode_ids.add(episode_id)

    return low_episode_ids


# Returns students that need improvement
def get_improvement_data(attempts_per_student):
    the_students = []  # that need improvement
    for student, best_attempts in attempts_per_student.items():
        episodes_of_concern = _check_attempts(best_attempts)
        if episodes_of_concern:
            areas = [messages.get_episode_title(ep_id) for ep_id in episodes_of_concern]
            areas_summary = ", ".join(areas)
            the_students.append(StudentInTrouble(student=student, areas=areas_summary))
    return the_students


def scoreboard_view(
    request, form, student_data, headers, level_headers, improvement_data
):
    database_episodes = level_selection.fetch_episode_data(False)

    return render(
        request,
        "game/scoreboard.html",
        context={
            "anchor": request.POST,  # Scroll to top of scoreboard section only if request is POST
            "form": form,
            "student_data": student_data,
            "headers": headers,
            "level_headers": level_headers,
            "total_points_header": TotalPointsHeader,
            "episodes": database_episodes,
            "improvement_data": improvement_data,
        },
    )


def scoreboard(request):
    """
    Renders a page with students' scores. A teacher can see the visible classes in
    their school. Student's view is restricted to their class if their teacher enabled
    the scoreboard for said class.
    """
    if not permissions.can_see_scoreboard(request.user):
        return render_no_permission_error(request)

    user = User(request.user.userprofile)
    users_classes = classes_for(user)
    all_episode_ids = list(range(1, 12))

    if user.is_independent_student():
        return render_no_permission_error(request)

    if request.POST:
        class_ids = set(map(int, request.POST.getlist("classes")))
        # Show all levels if the teacher doesn't select any
        episode_ids = (
            set(all_episode_ids)
            if "episodes" not in request.POST
            else set(map(int, request.POST.getlist("episodes")))
        )
    else:
        # Show no data on page load by default (if teacher)
        class_ids = ()
        episode_ids = ()

        # If user is a student, show data for their class and all episodes
        if user.is_student():
            class_ids = {user.student.class_field.id}
            episode_ids = set(all_episode_ids)

    if is_teacher_with_no_classes_assigned(user, users_classes):
        return render_no_permission_error(request)

    if not is_valid_request(user, class_ids):
        raise Http404

    form = ScoreboardForm(
        request.POST or None,
        classes=users_classes,
        initial={
            "classes": class_ids,
            "episodes": episode_ids,
        },  # Check selected checkboxes of each dropdown to cater for POST page load
    )

    classes = Class.objects.filter(id__in=class_ids)
    students = students_visible_to_user(user, classes)

    all_levels = []

    for episode_id in all_episode_ids:
        episode = Episode.objects.get(id=episode_id)
        all_levels += episode.levels

    attempts_per_student = {}

    for student in students:
        best_attempts = Attempt.objects.filter(
            level__in=all_levels, student=student, is_best_attempt=True
        ).select_related("level")
        attempts_per_student[student] = best_attempts

    student_data, headers, level_headers, levels_sorted = scoreboard_data(
        episode_ids, attempts_per_student
    )
    improvement_data = get_improvement_data(attempts_per_student)

    csv_export = "export" in request.POST

    if csv_export:
        return scoreboard_csv(student_data, levels_sorted, improvement_data)
    else:
        return scoreboard_view(
            request, form, student_data, headers, level_headers, improvement_data
        )


def render_no_permission_error(request):
    return renderError(
        request, messages.noPermissionTitle(), messages.noPermissionScoreboard()
    )


def is_teacher_with_no_classes_assigned(user, users_classes):
    return user.is_teacher() and len(users_classes) == 0


def success_response(form, student_data, headers):
    return form, student_data, headers, None


def classes_for(user):
    if user.is_teacher():
        if user.teacher.is_admin:
            teachers = Teacher.objects.filter(school=user.teacher.school)
            return Class.objects.filter(teacher__in=teachers)
        else:
            return Class.objects.filter(teacher=user.teacher)

    elif user.is_student():
        class_ = user.student.class_field
        return Class.objects.filter(id=class_.id)


def sorted_levels_by(level_ids):
    return sort_levels(Level.objects.filter(id__in=level_ids))


def are_classes_viewable_by_teacher(class_ids, user):
    teachers = Teacher.objects.filter(school=user.teacher.school)
    classes_in_teachers_school = Class.objects.filter(teacher__in=teachers).values_list(
        "id", flat=True
    )
    for class_id in class_ids:
        is_authorised = class_id in classes_in_teachers_school
        if not is_authorised:
            return False
    return True


def authorised_student_access(class_, class_ids):
    return len(class_ids) == 1 and next(iter(class_ids)) == class_


def students_visible_to_student(student):
    class_ = student.class_field
    if is_viewable(class_):
        return class_.students.filter(new_user__is_active=True).select_related(
            "class_field", "user__user"
        )
    else:
        return [student]


def students_visible_to_user(user, classes):
    if user.is_teacher():
        return students_of_classes(classes)
    elif user.is_student():
        student = user.student
        return students_visible_to_student(student)


def students_of_classes(classes):
    return Student.objects.filter(
        class_field__in=classes, new_user__is_active=True
    ).select_related("class_field", "user__user")


def is_valid_request(user, class_ids):
    if len(class_ids) == 0:
        return True
    elif user.is_teacher():
        return are_classes_viewable_by_teacher(class_ids, user)
    elif user.is_student():
        return authorised_student_access(user.student.class_field.id, class_ids)
    else:
        return False


def is_viewable(class_):
    return class_.classmates_data_viewable


def chop_miliseconds(delta):
    return delta - timedelta(microseconds=delta.microseconds)


class User(object):
    def __init__(self, profile):
        self.profile = profile
        if self.is_teacher():
            self.teacher = profile.teacher

        if self.is_student():
            self.student = profile.student

    def is_student(self):
        return (
            hasattr(self.profile, "student")
            and not self.profile.student.is_independent()
        )

    def is_teacher(self):
        return hasattr(self.profile, "teacher")

    def is_independent_student(self):
        return (
            hasattr(self.profile, "student") and self.profile.student.is_independent()
        )
